//
// Created by huli on 2020/7/20.
//

#ifndef NNXX_TESTS_CARRIER_BASE_H
#define NNXX_TESTS_CARRIER_BASE_H

#include "../error_code/error_code.h"
#include <glog/logging.h>
#include <thread>
#include <mutex>
#include "../tool/thread_condition.h"
#include "../dispatch/carrier_task.h"


//搬运器的基类, 普爱项目就是 电梯和中跑车和小跑车 三合一
class Carrier_base
{
public:
	//搬运器底层通信延时5000ms
#define	CARRIER_COMMUNICATION_OVER_TIME_MS	5000

	//搬运器状态,这个类的总状态,这里指工作任务流程
	enum Carrier_status
	{
		E_UNKNOW               	= 0,    //未知
		E_READY               	= 1,    //准备，待机
		E_BUSY					= 2, 	//工作正忙

		E_ONE_LEVEL_WORK		= 3,	//一级工作状态, 就是普通的移动任务
		E_ONE_LEVEL_OVER		= 4, 	//一级工作完成,
		E_TWO_LEVEL_WORK		= 5,	//二级工作状态, 就是紧急避让, 插队任务(会在执行一级任务的过程中, 插队并优先执行二级任务)
		E_TWO_LEVEL_OVER		= 6, 	//一级工作完成,
		E_THREE_LEVEL_WORK		= 7,	//三级工作任务, 就是锁定硬件资源, 不让插队, (除非急停或者取消任务)
		E_THREE_LEVEL_OVER		= 8, 	//一级工作完成,

		E_FAULT					= 100,	//故障
		E_DISCONNECT			= 101, 	//通信故障
	};

	//小跑车的夹杆
	enum Clamp_motion
	{
		E_CLAMP_NO_ACTION            	= 0,    //无动作.
	    E_CLAMP_TIGHT               	= 1,    //夹紧夹杆
	    E_CLAMP_LOOSE					= 2,	//松开夹杆
	};

	//中跑车和平层轨道的对接
	enum Joint_motion
	{
		E_JOINT_NO_ACTION            	= 0,    //无动作.
		E_JOINT_HOLD_OUT               	= 1,    //伸出对接，之后中跑车可以x轴移动，电梯不能Z轴移动
		E_JOINT_TAKE_BACK				= 2,	//收回对接，之后中跑车固定在电梯上不能X轴移动，电梯可以Z轴移动
	};

	//小跑车的位置,是否在中跑车上面
	enum Small_sports_car_motion
	{
		E_SMALL_SPORTS_NO_ACTION		= 0,    //无动作.
		E_SMALL_SPORTS_CAR_GET_AWAY 	= 1,    //小跑车离开中跑车
		E_SMALL_SPORTS_CAR_GET_BACK		= 2,	//小跑车返回中跑车
	};

	
	//指令完成状态, 搬运器答复指令, 返回任务完成的情况
	enum Respons_status
	{
		RESPONS_WORKING            	= 0,    //任务进行中
		RESPONS_OVER               	= 1,    //任务完成
		RESPONS_MINOR_ERROR			= 100,	//一般故障, 可恢复
		RESPONS_CRITICAL_ERROR		= 101,	//致命故障,不可恢复
	};

	//搬运器的硬件设备状态
	enum Device_status
	{
		DEVICE_UNKNOWN              = 0,    //设备未知
		DEVICE_READY                = 1,    //设备空闲(可以接受新的指令任务)
		DEVICE_WORKING				= 2,	//设备工作中
		DEVICE_EMERGENCY_STOP		= 3,	//设备急停
		DEVICE_UNSAFETY				= 4,	//设备不安全
		DEVICE_COLLISION			= 5,	//设备发生碰撞
		DEVICE_FAULT				= 6,	//设备故障
	};

	//搬运器的负载状态, 小跑车上面是否有车.
	enum Load_status
	{
		LOAD_UNKNOWN				= 0,	//负载未知
		HAVE_CAR               		= 1,    //有车
		NO_CAR            			= 2,    //没车
	};


public:
	Carrier_base();
	Carrier_base(const Carrier_base& other)= default;
	Carrier_base& operator =(const Carrier_base& other)= default;
	~Carrier_base();
public://API functions
	//搬运器 初始化
	virtual Error_manager carrier_base_init();
	//搬运器 反初始化
	virtual Error_manager carrier_base_uninit();

	//处理一级任务, 普通任务
	virtual Error_manager execute_one_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//处理二级任务, 插队任务
	virtual Error_manager execute_two_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//处理三级任务, 核心任务
	virtual Error_manager execute_three_level_task(std::shared_ptr<Carrier_task> p_carrier_task);
	//检查状态,是否正常运行
	virtual Error_manager check_status();

	//注意了, 调度任务允许同时接受多个任务
	//判断能否执行一级任务
	virtual Error_manager	check_one_level_task();
	//判断能否执行二级任务
	virtual Error_manager	check_two_level_task();
	//判断能否执行三级任务
	virtual Error_manager	check_three_level_task();

	//结束任务单，里面会根据任务的故障等级修正 任务单的状态
	virtual Error_manager end_task(std::shared_ptr<Carrier_task> carrier_task);
	//取消任务单，由发送方提前取消任务单
	virtual Error_manager cancel_task(std::shared_ptr<Carrier_task> carrier_task);

	//判断是否为待机，如果已经准备好，则可以执行任务。
	virtual bool is_ready();
public://get or set member variable
	Carrier_status get_carrier_status();

protected://member functions
	//执行外界任务的执行函数
	void execute_thread_fun();

	//执行线程工作函数, 正在执行任务单.
	virtual Error_manager execute_thread_working(std::shared_ptr<Carrier_task> p_carrier_task, Carrier_status & carrier_status);
	//执行线程工作函数, 已经完成任务单. 等待新的指令
	virtual Error_manager execute_thread_over(std::shared_ptr<Carrier_task> p_carrier_task, Carrier_status & carrier_status);


	//更新设备底层通信数据
	virtual Error_manager update_device_communication();

	//取消下发的指令
	virtual Error_manager cancel_command();


protected://member variable
	std::atomic<Carrier_status>			m_carrier_status;			//搬运器总状态, 控制任务流程
	int 								m_carrier_id;				//搬运器id, (楚天项目和单元号对应)




	//调度下发到plc
	std::string							m_request_key;				//请求唯一码, 用作识别
	//请求的目标坐标和动作
	float 								m_request_x;				//搬运器坐标x轴, 中跑车控制横向移动
	float 								m_request_y;				//搬运器坐标y轴, 小跑车控制纵向移动
	float 								m_request_z;				//搬运器坐标z轴, 电梯控制上下移动
	int 								m_request_space_id;			//搬运器空间位置的id.
	float 								m_request_y1;				//搬运器坐标y1轴, 小跑车控制纵向移动(前轮抓杆)
	float 								m_request_y2;				//搬运器坐标y2轴, 小跑车控制纵向移动(后轮抓杆)
	float 								m_request_wheelbase;		//搬运器小跑车的抓车杆前后轮距.
	Clamp_motion						m_request_clamp_motion;		//小跑车夹车杆. 0=无动作，1=夹紧，2=松开
	Joint_motion						m_request_joint_motion;		//电梯与X轴的横向轨道对接,0=无动作，
	//1=进行对接，之后中跑车可以x轴移动，电梯不能Z轴移动
	//2=松开对接，之后中跑车固定在电梯上不能X轴移动，电梯可以Z轴移动



	//plc反馈给调度
	std::string							m_respons_key;				//应答的唯一码, 用作识别
	Respons_status						m_respons_status;			//指令完成状态, 搬运器答复指令, 返回任务完成的情况
	//答复的实际坐标和动作
	float 								m_respons_x;				//搬运器坐标x轴, 中跑车控制横向移动
	float 								m_respons_y;				//搬运器坐标y轴, 小跑车控制纵向移动
	float 								m_respons_z;				//搬运器坐标z轴, 电梯控制上下移动
	int 								m_respons_space_id;			//搬运器空间位置的id.
	float 								m_respons_y1;				//搬运器坐标y1轴, 小跑车控制纵向移动(前轮抓杆)
	float 								m_respons_y2;				//搬运器坐标y2轴, 小跑车控制纵向移动(后轮抓杆)
	float 								m_respons_wheelbase;		//搬运器小跑车的抓车杆前后轮距.
	Clamp_motion						m_respons_clamp_motion;		//小跑车夹车杆. 0=无动作，1=夹紧，2=松开
	Joint_motion						m_respons_joint_motion;		//电梯与X轴的横向轨道对接,0=无动作，
	//1=进行对接，之后中跑车可以x轴移动，电梯不能Z轴移动
	//2=松开对接，之后中跑车固定在电梯上不能X轴移动，电梯可以Z轴移动




	//硬件状态, 目前只做显示判断
	std::chrono::system_clock::time_point	m_status_updata_time;	//状态更新时间点
	unsigned char						m_last_heartbeat;			//上一次的心跳
	//搬运器的设备状态数据,
	Device_status						m_device_status;			//搬运器的硬件设备状态
	Load_status							m_load_status;				//搬运器的负载状态, 小跑车上面是否有车.
	//搬运器的真实状态, 可能是路径中间的坐标
	float 								m_actual_x;					//搬运器坐标x轴, 中跑车控制横向移动
	float 								m_actual_y;					//搬运器坐标y轴, 小跑车控制纵向移动
	float 								m_actual_z;					//搬运器坐标z轴, 电梯控制上下移动
	float 								m_actual_y1;				//搬运器坐标y1轴, 小跑车控制纵向移动(前轮抓杆)
	float 								m_actual_y2;				//搬运器坐标y2轴, 小跑车控制纵向移动(后轮抓杆)
	Clamp_motion						m_actual_clamp_motion1;		//小跑车夹车杆. 0=无动作，1=夹紧，2=松开
	Clamp_motion						m_actual_clamp_motion2;		//小跑车夹车杆. 0=无动作，1=夹紧，2=松开
	Joint_motion						m_actual_joint_motion1;		//电梯与X轴的横向轨道对接,0=无动作，
	Joint_motion						m_actual_joint_motion2;		//电梯与X轴的横向轨道对接,0=无动作，
	Small_sports_car_motion				m_actual_small_sports_car_motion;	//小跑车的位置,是否在中跑车上面
	//搬运器设备的错误码
	unsigned char						m_carrier_error_code[50];	//搬运器设备的故障信息位
	unsigned char						m_carrier_warning_code[50];	//搬运器设备的警告信息位
	std::string							m_carrier_error_description;//搬运器设备的错误描述






	//任务执行线程
	std::thread*        				mp_execute_thread;   			//执行的线程指针，内存由本类管理
	Thread_condition					m_execute_condition;			//执行的条件变量

	std::shared_ptr<Carrier_task>		mp_carrier_one_level_task;		//搬运器的普通任务, (一般移动和等待指令, 允许被打断)
	std::shared_ptr<Carrier_task>		mp_carrier_two_level_task;		//搬运器的优先任务, (紧急避让和插队指令, 允许打断普通任务)
	std::shared_ptr<Carrier_task>		mp_carrier_three_level_task;	//搬运器的核心任务, (锁定位置进行抓车和放车, 最高任务, 不允许打断和被打断)



private:

};


#endif //NNXX_TESTS_CARRIER_BASE_H
